from typing import cast
from unittest import TestCase
from unittest.mock import call, patch, sentinel

from fred_api.dbreport.common_types_pb2 import (
    InputParameterDescription,
    JsonEncodedValue,
    OutputParameterDescription,
    ReportInfo,
    ReportName,
    Table,
)
from fred_api.dbreport.service_report_manager_grpc_pb2 import ManageReportsReply, ManageReportsRequest
from fred_api.dbreport.service_report_user_grpc_pb2 import ListReportsReply, RunReportReply, RunReportRequest
from frgal.utils import TestClientMixin
from google.protobuf.empty_pb2 import Empty

from reporter.client import ReportClient, ReportDecoder, ReportManagerClient
from reporter.report import Param, Report


class TestReportClient(TestClientMixin, ReportClient):
    """Testing version of ReportClient."""


class TestReportManagerClient(TestClientMixin, ReportManagerClient):
    """Testing version of ReportManagerClient."""


class ReportClientTestMixin:
    def get_in_param(self, type: str, name: str) -> InputParameterDescription:
        param = InputParameterDescription()
        param.data_type.value = type
        param.name.value = name
        return param

    def get_out_param(self, type: str, name: str) -> OutputParameterDescription:
        param = OutputParameterDescription()
        param.data_type.value = type
        param.name.value = name
        return param

    def get_field_desc(self, type: str, name: str) -> Table.FieldDescription:
        desc = Table.FieldDescription()
        desc.data_type.value = type
        desc.name = name
        return desc


class ReportClientTest(ReportClientTestMixin, TestCase):
    """Unittests for ReportClient."""

    def setUp(self):
        self.client = TestReportClient(sentinel.netloc)

    def test_list_reports_empty(self):
        reply = ListReportsReply()
        self.client.mock.return_value = reply

        result = self.client.list_reports()

        self.assertEqual(result, [])
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.DbReport.Api.ReportUser/list_reports", timeout=None)],
        )

    def test_list_reports(self):
        reply = ListReportsReply()
        report1 = ReportInfo()
        report1.name.value = "Report1"
        report2 = ReportInfo()
        report2.name.value = "Report2"
        reply.data.reports.extend([report1, report2])
        self.client.mock.return_value = reply

        result = self.client.list_reports()

        self.assertEqual(result, [Report("Report1", [], []), Report("Report2", [], [])])
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.DbReport.Api.ReportUser/list_reports", timeout=None)],
        )

    def test_list_reports_ambiguous(self):
        reply = ListReportsReply()
        report1 = ReportInfo()
        report1.name.value = "Unambiguous Report"
        report1.input_parameters.extend(
            [
                self.get_in_param("int", "param1"),
                self.get_in_param("text", "param2"),
            ]
        )
        report2 = ReportInfo()
        report2.name.value = "Ambiguous Report"
        report2.input_parameters.extend(
            [
                self.get_in_param("int", "param"),
                self.get_in_param("text", "param"),
            ]
        )
        reply.data.reports.extend([report1, report2])
        self.client.mock.return_value = reply

        result = self.client.list_reports()

        self.assertEqual(result, [Report("Unambiguous Report", [Param("param1", "int"), Param("param2", "text")], [])])
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.DbReport.Api.ReportUser/list_reports", timeout=None)],
        )

    def test_get_report(self):
        with patch.object(
            self.client, "list_reports", return_value=[Report("Another Report", [], []), Report("Report", [], [])]
        ):
            result = self.client.get_report("Report")
        self.assertEqual(cast(Report, result).name, "Report")

    def test_get_report_not_found(self):
        with patch.object(self.client, "list_reports", return_value=[]):
            result = self.client.get_report("Report")
        self.assertIsNone(result)

    def test_run_report(self):
        reply1 = RunReportReply()
        reply1.data.table.header.field_descriptions.extend(
            [
                self.get_field_desc("varchar", "fqdn"),
                self.get_field_desc("date", "exdate"),
            ]
        )
        reply1.data.table.rows.extend(
            [
                Table.JsonEncodedFields(value='["domain1.cz", "2020-01-01"]'),
                Table.JsonEncodedFields(value='["domain2.cz", "2019-01-01"]'),
            ]
        )
        reply2 = RunReportReply()
        reply2.data.table.header.field_descriptions.extend(
            [
                self.get_field_desc("varchar", "fqdn"),
                self.get_field_desc("date", "exdate"),
            ]
        )
        reply2.data.table.rows.extend(
            [
                Table.JsonEncodedFields(value='["domain3.cz", "2021-01-01"]'),
            ]
        )
        self.client.mock.return_value = [reply1, reply2]

        result = self.client.run_report("Report", "arg1", 42)

        self.assertEqual(
            list(result),
            [
                {"fqdn": "domain1.cz", "exdate": "2020-01-01"},
                {"fqdn": "domain2.cz", "exdate": "2019-01-01"},
                {"fqdn": "domain3.cz", "exdate": "2021-01-01"},
            ],
        )

        request = RunReportRequest()
        request.report_name.value = "Report"
        request.arguments.extend([JsonEncodedValue(value='"arg1"'), JsonEncodedValue(value="42")])
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(request, method="/Fred.DbReport.Api.ReportUser/run_report", timeout=None)],
        )


class ReportManagerClientTest(ReportClientTestMixin, TestCase):
    """Unittests for ReportManagerClient."""

    def setUp(self):
        self.client = TestReportManagerClient(sentinel.netloc)

    def test_manage_reports(self):
        reply = ManageReportsReply()
        reply.data.deleted.extend([ReportName(value=x) for x in ("Del1", "Del2")])
        new_report = ReportInfo()
        new_report.name.value = "Cr1"
        new_report.input_parameters.extend(
            [
                self.get_in_param("timestamp", "start"),
                self.get_in_param("text", "filter"),
            ]
        )
        reply.data.created.append(new_report)

        self.client.mock.return_value = reply

        result = self.client.manage_reports("some sql")

        self.assertEqual(
            result,
            {
                "created": [
                    Report(
                        "Cr1",
                        [
                            Param(name="start", type="timestamp"),
                            Param(name="filter", type="text"),
                        ],
                        [],
                    )
                ],
                "deleted": ["Del1", "Del2"],
            },
        )

        request = ManageReportsRequest()
        request.sql = "some sql"
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(request, method="/Fred.DbReport.Api.ReportManager/manage_reports", timeout=None)],
        )


class ReportDecoderTest(ReportClientTestMixin, TestCase):
    """Unittests for ReportDecoder."""

    def setUp(self):
        self.decoder = ReportDecoder()

    def test_decode_report_info(self):
        report = ReportInfo()
        report.name.value = "Report"
        report.input_parameters.extend(
            [
                self.get_in_param("timestamp", "start"),
                self.get_in_param("text", "filter"),
            ]
        )
        report.output_parameters.extend(
            [
                self.get_out_param("integer", "count"),
                self.get_out_param("text", "name"),
            ]
        )

        self.assertEqual(
            self.decoder.decode(report),
            Report(
                "Report",
                [Param(name="start", type="timestamp"), Param(name="filter", type="text")],
                [Param(name="count", type="integer"), Param(name="name", type="text")],
            ),
        )
