from decimal import Decimal
from typing import Iterator
from unittest import TestCase
from unittest.mock import sentinel

from reporter.report import AmbiguousColumnNames, Param, Report, ReportResult


class ParamTest(TestCase):
    def test_repr(self):
        self.assertEqual(repr(Param(name="count", type="int")), "Param<count: int>")

    def test_eq(self):
        self.assertEqual(
            Param(name=sentinel.param_name, type=sentinel.param_type),
            Param(name=sentinel.param_name, type=sentinel.param_type),
        )
        self.assertNotEqual(
            Param(name=sentinel.param_name, type=sentinel.param_type),
            Param(name=sentinel.param_name2, type=sentinel.param_type),
        )
        self.assertNotEqual(
            Param(name=sentinel.param_name, type=sentinel.param_type),
            Param(name=sentinel.param_name, type=sentinel.param_type2),
        )
        self.assertNotEqual(
            Param(name=sentinel.param_name, type=sentinel.param_type),
            Param(name=sentinel.param_name2, type=sentinel.param_type2),
        )
        self.assertNotEqual(
            Param(name=sentinel.param_name, type=sentinel.param_type),
            "Param",
        )

    def test_to_dict(self):
        self.assertEqual(
            Param(name=sentinel.param_name, type=sentinel.param_type).to_dict(),
            {"name": sentinel.param_name, "type": sentinel.param_type},
        )


class ReportTest(TestCase):
    def test_to_dict(self):
        self.assertEqual(
            Report(sentinel.report_name, [], []).to_dict(),
            {"name": sentinel.report_name, "input_parameters": [], "output_parameters": []},
        )
        self.assertEqual(
            Report(
                sentinel.report_name,
                [Param(sentinel.param_name, sentinel.param_type)],
                [Param(sentinel.param_name2, sentinel.param_type2)],
            ).to_dict(),
            {
                "name": sentinel.report_name,
                "input_parameters": [{"name": sentinel.param_name, "type": sentinel.param_type}],
                "output_parameters": [{"name": sentinel.param_name2, "type": sentinel.param_type2}],
            },
        )

    def test_eq(self):
        self.assertEqual(
            Report(sentinel.report_name, [], []),
            Report(sentinel.report_name, [], []),
        )
        self.assertEqual(
            Report(sentinel.report_name, [Param(sentinel.param_name, sentinel.param_type)], []),
            Report(sentinel.report_name, [Param(sentinel.param_name, sentinel.param_type)], []),
        )
        self.assertEqual(
            Report(sentinel.report_name, [], [Param(sentinel.param_name, sentinel.param_type)]),
            Report(sentinel.report_name, [], [Param(sentinel.param_name, sentinel.param_type)]),
        )
        self.assertNotEqual(
            Report(sentinel.report_name, [], []),
            Report(sentinel.report_name2, [], []),
        )
        self.assertNotEqual(
            Report(sentinel.report_name, [Param(sentinel.param_name2, sentinel.param_type)], []),
            Report(sentinel.report_name, [Param(sentinel.param_name, sentinel.param_type)], []),
        )
        self.assertNotEqual(
            Report(sentinel.report_name, [], [Param(sentinel.param_name, sentinel.param_type)]),
            Report(sentinel.report_name, [], [Param(sentinel.param_name, sentinel.param_type2)]),
        )
        self.assertNotEqual(
            Report(sentinel.report_name, [], []),
            "Report",
        )

    def test_is_unambiguous(self):
        self.assertTrue(Report(sentinel.report_name, [], []).is_unambiguous)
        self.assertTrue(
            Report(
                sentinel.report_name,
                [
                    Param(sentinel.param_name, sentinel.param_type),
                    Param(sentinel.param_name2, sentinel.param_type),
                ],
                [],
            ).is_unambiguous
        )
        self.assertTrue(
            Report(
                sentinel.report_name,
                [],
                [
                    Param(sentinel.param_name, sentinel.param_type),
                    Param(sentinel.param_name2, sentinel.param_type),
                ],
            ).is_unambiguous
        )
        self.assertFalse(
            Report(
                sentinel.report_name,
                [
                    Param(sentinel.param_name, sentinel.param_type),
                    Param(sentinel.param_name, sentinel.param_type),
                ],
                [],
            ).is_unambiguous
        )
        self.assertFalse(
            Report(
                sentinel.report_name,
                [],
                [
                    Param(sentinel.param_name, sentinel.param_type),
                    Param(sentinel.param_name, sentinel.param_type),
                ],
            ).is_unambiguous
        )


class ReportResultTest(TestCase):
    def setUp(self):
        self.stream = iter(
            [
                {
                    "header": [
                        {"data_type": "varchar", "name": "fqdn"},
                        {"data_type": "date", "name": "exdate"},
                        {"data_type": "numeric", "name": "rating"},
                    ],
                    "rows": ['["domain1.cz", "2020-01-01", 3.14]', '["domain2.cz", "2019-01-01", 45.0000005]'],
                },
                {
                    "header": [
                        {"data_type": "varchar", "name": "fqdn"},
                        {"data_type": "date", "name": "exdate"},
                        {"data_type": "numeric", "name": "rating"},
                    ],
                    "rows": ['["domain3.cz", "2021-01-01", 10.0]'],
                },
            ]
        )

    def test_result(self):
        result = ReportResult(stream=self.stream)
        self.assertIsInstance(result, Iterator)
        self.assertEqual(
            list(result),
            [
                {"fqdn": "domain1.cz", "exdate": "2020-01-01", "rating": Decimal("3.14")},
                {"fqdn": "domain2.cz", "exdate": "2019-01-01", "rating": Decimal("45.0000005")},
                {"fqdn": "domain3.cz", "exdate": "2021-01-01", "rating": Decimal("10.0")},
            ],
        )
        # when we iterate over the result again, it is already exhausted
        self.assertEqual(list(result), [])

    def test_ambiguous(self):
        stream = iter(
            [
                {
                    "header": [{"data_type": "varchar", "name": "fqdn"}, {"data_type": "text", "name": "fqdn"}],
                    "rows": ['["domain1.cz", "2020-01-01"]', '["domain2.cz", "2019-01-01"]'],
                }
            ]
        )
        with self.assertRaises(AmbiguousColumnNames):
            ReportResult(stream=stream)

    def test_empty(self):
        stream: Iterator = iter([])
        self.assertEqual(list(ReportResult(stream=stream)), [])
